"""
Modified from https://github.com/pytorch/vision/blob/main/torchvision/extension.py
"""

import ctypes
import importlib
import os
import sys
from warnings import warn

import torch


def _get_extension_path(lib_name):
    lib_dir = os.path.dirname(__file__)
    if os.name == 'nt':
        # Register the main library location on the default DLL path
        import ctypes
        import sys

        kernel32 = ctypes.WinDLL('kernel32.dll', use_last_error=True)
        with_load_library_flags = hasattr(kernel32, 'AddDllDirectory')
        prev_error_mode = kernel32.SetErrorMode(0x0001)

        if with_load_library_flags:
            kernel32.AddDllDirectory.restype = ctypes.c_void_p

        if sys.version_info >= (3, 8):
            os.add_dll_directory(lib_dir)
        elif with_load_library_flags:
            res = kernel32.AddDllDirectory(lib_dir)
            if res is None:
                err = ctypes.WinError(ctypes.get_last_error())
                err.strerror += f' Error adding "{lib_dir}" to the DLL directories.'
                raise err

        kernel32.SetErrorMode(prev_error_mode)

    loader_details = (importlib.machinery.ExtensionFileLoader, importlib.machinery.EXTENSION_SUFFIXES)

    extfinder = importlib.machinery.FileFinder(lib_dir, loader_details)
    ext_specs = extfinder.find_spec(lib_name)
    if ext_specs is None:
        raise ImportError

    return ext_specs.origin


_HAS_OPS = False


def _has_ops():
    return False


try:
    # On Windows Python-3.8.x has `os.add_dll_directory` call,
    # which is called to configure dll search path.
    # To find cuda related dlls we need to make sure the
    # conda environment/bin path is configured Please take a look:
    # https://stackoverflow.com/questions/59330863/cant-import-dll-module-in-python
    # Please note: if some path can't be added using add_dll_directory we simply ignore this path
    if os.name == 'nt' and (3, 8) <= sys.version_info < (3, 9):
        env_path = os.environ['PATH']
        path_arr = env_path.split(';')
        for path in path_arr:
            if os.path.exists(path):
                try:
                    os.add_dll_directory(path)  # type: ignore[attr-defined]
                except Exception:
                    pass

    lib_path = _get_extension_path('_C')
    torch.ops.load_library(lib_path)
    _HAS_OPS = True


    def _has_ops():  # noqa: F811
        return True

except (ImportError, OSError):
    pass


def _assert_has_ops():
    if not _has_ops():
        raise RuntimeError(
            'Couldn\'t load custom C++ ops. Recompile C++ extension with:\n'
            '\tpython setup.py build_ext --inplace'
        )


def _check_cuda_version(minor=True):
    """
    Make sure that CUDA versions match between the pytorch install and tvdcn install

    Args:
        minor (bool): If ``False``, ignore minor version difference.
         Defaults to ``True``.
    """
    if not _HAS_OPS:
        return -1
    from torch.version import cuda as torch_version_cuda

    _version = torch.ops.tvdcn._cuda_version()
    if _version != -1 and torch_version_cuda is not None:
        ext_version = str(_version)
        if int(ext_version) < 10000:
            ext_major = int(ext_version[0])
            ext_minor = int(ext_version[2])
        else:
            ext_major = int(ext_version[0:2])
            ext_minor = int(ext_version[3])
        t_version = torch_version_cuda.split('.')
        t_major = int(t_version[0])
        t_minor = int(t_version[1])
        if t_major != ext_major or (minor and t_minor != ext_minor):
            raise RuntimeError(
                'Detected that PyTorch and Extension were compiled with different CUDA versions. '
                f'PyTorch has CUDA Version={t_major}.{t_minor} and '
                f'Extension has CUDA Version={ext_major}.{ext_minor}. '
                'Please reinstall the Extension that matches your PyTorch install.'
            )
    return _version


def _load_library(lib_name):
    lib_path = _get_extension_path(lib_name)
    # On Windows Python-3.8+ has `os.add_dll_directory` call,
    # which is called from _get_extension_path to configure dll search path
    # Condition below adds a workaround for older versions by
    # explicitly calling `LoadLibraryExW` with the following flags:
    #  - LOAD_LIBRARY_SEARCH_DEFAULT_DIRS (0x1000)
    #  - LOAD_LIBRARY_SEARCH_DLL_LOAD_DIR (0x100)
    if os.name == 'nt' and sys.version_info < (3, 8):
        _kernel32 = ctypes.WinDLL('kernel32.dll', use_last_error=True)
        if hasattr(_kernel32, 'LoadLibraryExW'):
            _kernel32.LoadLibraryExW(lib_path, None, 0x00001100)
        else:
            warn('LoadLibraryExW is missing in kernel32.dll')

    torch.ops.load_library(lib_path)


_check_cuda_version(False)


###################
# Exposed functions
###################
def has_ops():
    """
    Check if C++ extension is successfully compiled.
    """
    return _HAS_OPS


def cuda_version():
    """
    Get compiled Cuda version.
    """
    if not _HAS_OPS:
        return -1
    return torch.ops.tvdcn._cuda_version()


def with_cuda():
    """
    Check if C++ extension is compiled with Cuda.
    """
    return cuda_version() != -1



import torch
from torch import Tensor
import torch.nn as nn

__all__ = [
    'mask_sigmoid',
    'MaskSigmoid',
]


def mask_sigmoid(mask: Tensor, scale: float = 2.0) -> Tensor:
    r"""
    Applies scaled sigmoid activation on mask. The default scale is set to 2 so that initial
    values when ``conv_mask`` is initialized with zeros is 1.

    References:
        https://github.com/msracver/Deformable-ConvNets/blob/master/DCNv2_op/example_symbol.py

    Args:
        mask (Tensor[batch_size, mask_groups * kernel_area, *out_shape]): modulation masks
            to be multiplied with each output of convolution kernel.
        scale (float): positive scaling of the activation output.
    """
    if scale <= 0:
        raise ValueError('scale must be positive. Got scale={}'.format(scale))
    return torch.sigmoid(mask) * scale


class MaskSigmoid(nn.Module):
    """
    See :func:`mask_sigmoid`.
    """

    def __init__(self, scale: float = 2.0):
        super().__init__()
        self.scale = float(scale)

    def forward(self, mask: Tensor) -> Tensor:
        return mask_sigmoid(mask, self.scale)



import torch
import torchvision

import tvdcn
from utils.deform_conv_test_args import DeformConvTestArgs


def test_deform_conv2d(dtype=torch.float64,
                       device='cuda',
                       with_mask=True):
    torch.manual_seed(12)
    args = DeformConvTestArgs(dim=2,
                              offset_groups=1,
                              mask_groups=1,
                              dtype=dtype,
                              device=device)
    print(args)

    args.zero_grad()
    tvdcn_output = tvdcn.ops.deform_conv2d(args.input,
                                           args.weight,
                                           args.offset,
                                           args.mask if with_mask else None,
                                           args.bias,
                                           args.stride,
                                           args.padding,
                                           args.dilation,
                                           args.groups)
    tvdcn_output.sum().backward()
    tvdcn_input_grad = args.input.grad.clone()
    tvdcn_weight_grad = args.weight.grad.clone()
    tvdcn_offset_grad = args.offset.grad.clone()
    tvdcn_mask_grad = args.mask.grad.clone() if with_mask else torch.zeros([0])
    tvdcn_bias_grad = args.bias.grad.clone()

    args.zero_grad()
    torchvision_output = torchvision.ops.deform_conv2d(args.input,
                                                       args.offset,
                                                       args.weight,
                                                       args.bias,
                                                       args.stride,
                                                       args.padding,
                                                       args.dilation,
                                                       args.mask if with_mask else None)
    torchvision_output.sum().backward()
    torchvision_input_grad = args.input.grad.clone()
    torchvision_weight_grad = args.weight.grad.clone()
    torchvision_offset_grad = args.offset.grad.clone()
    torchvision_mask_grad = args.mask.grad.clone() if with_mask else torch.zeros([0])
    torchvision_bias_grad = args.bias.grad.clone()

    print('output_diff:', tvdcn_output - torchvision_output, sep='\n')
    assert torch.allclose(tvdcn_output, torchvision_output)

    print('\ngrad_diff:')
    print(f'\tinput_grad:         {torch.allclose(tvdcn_input_grad, torchvision_input_grad)}')
    assert torch.allclose(tvdcn_input_grad, torchvision_input_grad)
    print(f'\tweight_grad:        {torch.allclose(tvdcn_weight_grad, torchvision_weight_grad)}')
    assert torch.allclose(tvdcn_weight_grad, torchvision_weight_grad)
    print(f'\toffset_grad:        {torch.allclose(tvdcn_offset_grad, torchvision_offset_grad)}')
    assert torch.allclose(tvdcn_offset_grad, torchvision_offset_grad)
    print(f'\tmask_grad:          {torch.allclose(tvdcn_mask_grad, torchvision_mask_grad)}')
    assert torch.allclose(tvdcn_mask_grad, torchvision_mask_grad)
    print(f'\tbias_grad:          {torch.allclose(tvdcn_bias_grad, torchvision_bias_grad)}')
    assert torch.allclose(tvdcn_bias_grad, torchvision_bias_grad)


if __name__ == '__main__':
    test_deform_conv2d()
